unit NumberConvert;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type

  TNumberName = record
    Number: integer;
    NameMasculine,  // If it stands in front of masculine gender.
    NameFeminine    // If it stands in front of feminine gender. If empty, NameMasculine is used (meaning feminine and masculine name is the same).
    //,NameNeuter // Unused for now.
     : string;
  end;

  TNumberNames = array of TNumberName;

  TGender = (gMasculine,gFeminine,rNeuter);

  // If Name5 or Name 2 is not set, then Name1 is used. If Name1 is not set, Name1Ignored is used.

  TNumberOrder = record
    Order: integer; // Number of zeros (exponent of the number system base).
    Gender: TGender;   // Male or female (neutral is not used for now).
    Name1Ignored: string; // If prefix "one" is excluded.
    Name1: string;
    Name2: string;
    Name5: string;
  end;

  TNumberOrders = array of TNumberOrder;

  TNumberLanguage = (nlEnglishUS,nlEnglishBritish,nlSerbian,nlSerbianFont,nlCustom);

  TNumberToText = class(TComponent)
  private
    FLanguage: TNumberLanguage;
    FSeparator: string;
    FIgnoreOne: boolean;
    FDecimalSeparator: string;
    FMinusName: string;
    FNumber: Extended;
    function GetText: string;
    procedure SetText(const Value: string);
  protected
    procedure SetLanguage(const Value: TNumberLanguage); virtual;
  public
    NumberNames: TNumberNames;
    NumberOrders: TNumberOrders;
    constructor Create(AOwner: TComponent); override;
  published
    property DecimalSeparator: string read FDecimalSeparator write FDecimalSeparator;
    property IgnoreOne: boolean read FIgnoreOne write FIgnoreOne;
    property Language: TNumberLanguage read FLanguage write SetLanguage;
    property MinusName: string read FMinusName write FMinusName;
    property Number: Extended read FNumber write FNumber;
    property Separator: string read FSeparator write FSeparator;
    property Text: string read GetText write SetText stored false;
  end;

  TNumberToRoman = class(TComponent)
  private
    FNumber: Cardinal;
    function GetRomanNumber: string;
    procedure SetRomanNumber(const Value: string);
  published
    property Number: Cardinal read FNumber write FNumber;
    property RomanNumber: string read GetRomanNumber write SetRomanNumber stored false;
  end;

  // TODO: RelatedComponent - automatically update Caption or Text.

procedure Register;
function CardToRoman(a: Cardinal): string;
function TrimRightZeros(const x: Int64): Int64;



implementation

uses
  Math;

{$R *.dcr}

procedure Register;
begin
  RegisterComponents('BConvert', [TNumberToText,TNumberToRoman]);
end;



// Converts 'a' into string containing roman form of that number.
// NOTE: Special cases, when I is placed in front of some
// digits are treated separately - it's simplest.
function CardToRoman(a: Cardinal): string;
const
  br=13; // Broj cifara.
type
  oznakatip = array[0..br-1] of string;
  vrednosttip = array[0..br-1] of Cardinal;
const (* I V X L C D M *)
  oznaka: oznakatip = (    'I','IV','V','IX','X','XL','L','XC','C','CD','D','CM','M' );
  vrednost: vrednosttip = (1,  4,   5,  9,   10, 40,  50, 90,  100,400, 500,900, 1000);
var
  //cch: Cardinal; // Tekuci karakter u 'r'.
  indeks: 0..br-1;

  // Vraca indeks rimske cifre koja je najveca od onih koje su manje od 'max'.
  function NajvecaManjaIliJednaka(const max: Cardinal): Cardinal;
  var
    curr: Cardinal;
  begin
    curr := Low(vrednost);
    while curr < High(vrednost) do  // Ne mora <=.
    begin
      if vrednost[curr+1] <= max then
        Inc(curr)
      else
      begin
        if vrednost[curr] <= max then
        begin
          result := curr;
          exit
        end
        else
          raise Exception.Create('NajvecaManjaIliJednaka(): ''max'' je premali.');
      end
    end;
    result := High(vrednost)
  end;

begin
  result := '';
  while a > 0 do
  begin
    indeks := NajvecaManjaIliJednaka(a);
    Dec(a,vrednost[indeks]);
    result := result + oznaka[indeks];
  end
end;

// Get rid of trailing zeros.
function TrimRightZeros(const x: Int64): Int64;
begin
  result := x;
  while (result <> 0) and ((result mod 10) = 0) do
    result := result div 10
end;

// Internal number-to-text conversion routine.
function NumberToText(
  Brojevi: TNumberNames;
  Redovi: TNumberOrders;
  const MinusName: string;
  Broj: Int64;
  const DecimalSeparatorName: string;
  Fraction: Int64 = 0;
  FractionPos: integer = 0;
  const Separator: string = ' ';
  const ZanemariJedan: boolean = true
): string;

  // Dodaje 'S1' na 'S' i umece separator izmedju (ali samo ako je potrebno).
  function AppendSepStr(const S,S1: string): string;
  begin
    result := S;
    if S1 = '' then
      exit;
    if not (S = '') then
      result := result + Separator;
    result := result + S1
  end;

  function BS(Broj: Int64; const Rod: TGender; const Zanemari: boolean): string;
  var

    i: integer;
    tmp,vrednost_reda: Int64;

    min_kolicnik,max_ostatak: Int64; // // U pretrazi redova.
    max_broj: Int64;  // // U pretrazi brojeva.
    Indeks: integer; // Indeks nadjenog reda ili broja.

    TmpString: string;

  begin

    result := '';

    // Prvo nalazimo maksimalni red (on ce proizvesti minimalni kolicnik veci od 0).
    Indeks := -1;
    min_kolicnik := Broj;
    max_ostatak := -1;

    for i := Low(Redovi) to High(Redovi) do
    begin
      vrednost_reda := Round(IntPower(10,Redovi[i].Order));
      tmp := Broj div vrednost_reda;
      if (tmp > 0) and (tmp < min_kolicnik) then
      begin
        min_kolicnik := tmp;
        max_ostatak := Broj mod vrednost_reda;
        Indeks := i;
      end
    end;

    // Ako je red nadjen, rekurzija.
    if Indeks >= 0 then
    begin
      // Nalazimo odgovarajucu verziju naziva Reda.
      if ((min_kolicnik mod 100) >= 5) and ((min_kolicnik mod 100) <= 20) then
        TmpString := Redovi[Indeks].Name5
      else
        case min_kolicnik mod 10 of
          1:
          begin
            if (min_kolicnik = 1) and Zanemari then
              TmpString := Redovi[Indeks].Name1Ignored
            else
              TmpString := Redovi[Indeks].Name1;
          end;
          2,3,4:
            TmpString := Redovi[Indeks].Name2;
          else
            TmpString := Redovi[Indeks].Name5
        end;

      if TmpString = '' then
        TmpString := Redovi[Indeks].Name1;

      if TmpString = '' then
        TmpString := Redovi[Indeks].Name1Ignored;

      if TmpString = '' then
        raise Exception.Create('NumberToText() Name of the order ' + IntToStr(Redovi[Indeks].Order) + ' is not set.');

      // Rekurzivno obradjemo kolicnik i ostatak (red je izmedju).
      if not ((min_kolicnik = 1) and Zanemari) then
        result := AppendSepStr(result,BS(min_kolicnik,Redovi[Indeks].Gender,Zanemari));
      result := AppendSepStr(result,TmpString);
      result := AppendSepStr(result,BS(max_ostatak,gMasculine,false));

    end

    // Ako red nije nadjen, posmatramo Broj kao zbir elemenata niza Brojevi.
    else
    begin
      while Broj > 0 do
      begin

        // Nalazimo maksimalni broj.
        Indeks := -1;
        max_broj := 0;
        for i := Low(Brojevi) to High(Brojevi) do
        begin
          tmp := Brojevi[i].Number;
          if (tmp > max_broj) and (tmp <= Broj) then
          begin
            max_broj := tmp;
            Indeks := i;
          end
        end;

        if max_broj = 0 then
          exit;

        Broj := Broj - max_broj;

        // A zatim njegov naziv u odgovarajucem rodu nastavljamo na rezultujuci
        // string.
        TmpString := '';
        case Rod of
          gMasculine:
            TmpString := Brojevi[Indeks].NameMasculine;
          gFeminine:
          begin
            TmpString := Brojevi[Indeks].NameFeminine;
            if TmpString = '' then
              TmpString := Brojevi[Indeks].NameMasculine
          end
        end; // case Rod
        Assert(not (TmpString = ''));
        // Ovo nam cak ni ne treba.
        //result := result + TmpString;
        result := AppendSepStr(result,TmpString);

      end // while Broj > 0

    end // else; if Indeks >= 0

  end;

  // Finds masculine name of number 'N'.
  function FindNameMasc(const N: Int64): string;
  var
    i: integer;
  begin
    result := '';
    for i := Low(Brojevi) to High(Brojevi) do
      if Brojevi[i].Number = N then
      begin
        result := Brojevi[i].NameMasculine
      end
  end;

  function FindZeroName: string;
  begin
    result := FindNameMasc(0)
  end;

var
  TmpFrac: string;
  ZeroName: string;

begin
  if Broj = 0 then
    result := FindZeroName
    //result := 'nula' // TODO: Search for zero in number names.
  else
  begin
    if Broj < 0 then
    begin
      result := MinusName;
      Broj := -Broj
    end
    else
      result := '';
    result := AppendSepStr(result,BS(Broj,gMasculine,ZanemariJedan));
  end;

  // Zatim obradjijemo decimalni deo.

  if Fraction < 0 then
    raise Exception.Create('NumberToText() Fraction can''t be less then zero.')
  else if Fraction > 0 then
  begin
    TmpFrac := '';
    while Fraction > 0 do
    begin
      TmpFrac := AppendSepStr(FindNameMasc(Fraction mod 10),TmpFrac);
      Fraction := Fraction div 10
    end;

    ZeroName := FindZeroName;
    while FractionPos > 0 do
    begin
      TmpFrac := AppendSepStr(ZeroName,TmpFrac);
      Dec(FractionPos)
    end;

    TmpFrac := AppendSepStr(DecimalSeparatorName,TmpFrac);
    result := AppendSepStr(result,TmpFrac);
  end

end;



{ TNumberToText }

constructor TNumberToText.Create(AOwner: TComponent);
begin
  inherited;
  Language := nlEnglishUS; // We don't set FLanguage because we want to trigger SetLanguage.
  FSeparator := ' ';
  FIgnoreOne := false;
  FNumber := 0
end;

procedure TNumberToText.SetLanguage(const Value: TNumberLanguage);
begin
  FLanguage := Value;
  case FLanguage of
  nlEnglishUS,nlEnglishBritish:
    begin
      FDecimalSeparator := 'point';
      FMinusName := 'minus';

      // First, we initialize number names.

      SetLength(NumberNames,28);

      with NumberNames[0] do
      begin
        Number := 0;
        NameMasculine := 'zero';
        NameFeminine := ''
      end;

      with NumberNames[1] do
      begin
        Number := 1;
        NameMasculine := 'one';
        NameFeminine := ''
      end;

      with NumberNames[2] do
      begin
        Number := 2;
        NameMasculine := 'two';
        NameFeminine := ''
      end;

      with NumberNames[3] do
      begin
        Number := 3;
        NameMasculine := 'three';
        NameFeminine := ''
      end;

      with NumberNames[4] do
      begin
        Number := 4;
        NameMasculine := 'four';
        NameFeminine := ''
      end;

      with NumberNames[5] do
      begin
        Number := 5;
        NameMasculine := 'five';
        NameFeminine := ''
      end;

      with NumberNames[6] do
      begin
        Number := 6;
        NameMasculine := 'six';
        NameFeminine := ''
      end;

      with NumberNames[7] do
      begin
        Number := 7;
        NameMasculine := 'seven';
        NameFeminine := ''
      end;

      with NumberNames[8] do
      begin
        Number := 8;
        NameMasculine := 'eight';
        NameFeminine := ''
      end;

      with NumberNames[9] do
      begin
        Number := 9;
        NameMasculine := 'nine';
        NameFeminine := ''
      end;

      with NumberNames[10] do
      begin
        Number := 10;
        NameMasculine := 'ten';
        NameFeminine := ''
      end;

      with NumberNames[11] do
      begin
        Number := 11;
        NameMasculine := 'eleven';
        NameFeminine := ''
      end;

      with NumberNames[12] do
      begin
        Number := 12;
        NameMasculine := 'twelve';
        NameFeminine := ''
      end;

      with NumberNames[13] do
      begin
        Number := 13;
        NameMasculine := 'thirteen';
        NameFeminine := ''
      end;

      with NumberNames[14] do
      begin
        Number := 14;
        NameMasculine := 'fourteen';
        NameFeminine := ''
      end;

      with NumberNames[15] do
      begin
        Number := 15;
        NameMasculine := 'fifteen';
        NameFeminine := ''
      end;

      with NumberNames[16] do
      begin
        Number := 16;
        NameMasculine := 'sixteen';
        NameFeminine := ''
      end;

      with NumberNames[17] do
      begin
        Number := 17;
        NameMasculine := 'seventeen';
        NameFeminine := ''
      end;

      with NumberNames[18] do
      begin
        Number := 18;
        NameMasculine := 'eighteen';
        NameFeminine := ''
      end;

      with NumberNames[19] do
      begin
        Number := 19;
        NameMasculine := 'nineteen';
        NameFeminine := ''
      end;

      with NumberNames[20] do
      begin
        Number := 20;
        NameMasculine := 'twenty';
        NameFeminine := ''
      end;

      with NumberNames[21] do
      begin
        Number := 30;
        NameMasculine := 'thirty';
        NameFeminine := ''
      end;

      with NumberNames[22] do
      begin
        Number := 40;
        NameMasculine := 'forty';
        NameFeminine := ''
      end;

      with NumberNames[23] do
      begin
        Number := 50;
        NameMasculine := 'fifty';
        NameFeminine := ''
      end;

      with NumberNames[24] do
      begin
        Number := 60;
        NameMasculine := 'sixty';
        NameFeminine := ''
      end;

      with NumberNames[25] do
      begin
        Number := 70;
        NameMasculine := 'seventy';
        NameFeminine := ''
      end;

      with NumberNames[26] do
      begin
        Number := 80;
        NameMasculine := 'eighty';
        NameFeminine := ''
      end;

      with NumberNames[27] do
      begin
        Number := 90;
        NameMasculine := 'ninety';
        NameFeminine := ''
      end;

      // Then we initialize order names.

      SetLength(NumberOrders,6);

      with NumberOrders[0] do
      begin
        Order := 2;
        Gender := gFeminine;
        Name1Ignored := 'hundred';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

      with NumberOrders[1] do
      begin
        Order := 3;
        Gender := gFeminine;
        Name1Ignored := 'thousand';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

      with NumberOrders[2] do
      begin
        Order := 6;
        Gender := gMasculine;
        Name1Ignored := 'million';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

      with NumberOrders[3] do
      begin
        Order := 9;
        Gender := gFeminine;
        if FLanguage = nlEnglishUS then
          Name1Ignored := 'billion'
        else
          Name1Ignored := 'milliard';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

      with NumberOrders[4] do
      begin
        Order := 12;
        Gender := gMasculine;
        if FLanguage = nlEnglishUS then
          Name1Ignored := 'trillion'
        else
          Name1Ignored := 'billion';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

      with NumberOrders[5] do
      begin
        Order := 15;
        Gender := gFeminine;
        if FLanguage = nlEnglishUS then
          Name1Ignored := 'quadrillion'
        else
          Name1Ignored := 'billiard';
        Name1 := '';
        Name2 := '';
        Name5 := '';
      end;

    end;

  nlSerbian,nlSerbianFont:
    begin
      FDecimalSeparator := 'zarez';
      FMinusName := 'minus';

      // Prvo inicijalizujemo brojeve.

      SetLength(NumberNames,37);

      with NumberNames[0] do
      begin
        Number := 0;
        NameMasculine := 'nula';
        NameFeminine := ''
      end;

      with NumberNames[1] do
      begin
        Number := 1;
        NameMasculine := 'jedan';
        NameFeminine := 'jedna'
      end;

      with NumberNames[2] do
      begin
        Number := 2;
        NameMasculine := 'dva';
        NameFeminine := 'dve'
      end;

      with NumberNames[3] do
      begin
        Number := 3;
        NameMasculine := 'tri';
        NameFeminine := ''
      end;

      with NumberNames[4] do
      begin
        Number := 4;
        if FLanguage = nlSerbian then
          NameMasculine := 'cetiri'
        else
          NameMasculine := 'etiri';
        NameFeminine := ''
      end;

      with NumberNames[5] do
      begin
        Number := 5;
        NameMasculine := 'pet';
        NameFeminine := ''
      end;

      with NumberNames[6] do
      begin
        Number := 6;
        if FLanguage = nlSerbian then
          NameMasculine := 'sest'
        else
          NameMasculine := 'est';
        NameFeminine := ''
      end;

      with NumberNames[7] do
      begin
        Number := 7;
        NameMasculine := 'sedam';
        NameFeminine := ''
      end;

      with NumberNames[8] do
      begin
        Number := 8;
        NameMasculine := 'osam';
        NameFeminine := ''
      end;

      with NumberNames[9] do
      begin
        Number := 9;
        NameMasculine := 'devet';
        NameFeminine := ''
      end;

      with NumberNames[10] do
      begin
        Number := 10;
        NameMasculine := 'deset';
        NameFeminine := ''
      end;

      with NumberNames[11] do
      begin
        Number := 11;
        NameMasculine := 'jedanaest';
        NameFeminine := ''
      end;

      with NumberNames[12] do
      begin
        Number := 12;
        NameMasculine := 'dvanaest';
        NameFeminine := ''
      end;

      with NumberNames[13] do
      begin
        Number := 13;
        NameMasculine := 'trinaest';
        NameFeminine := ''
      end;

      with NumberNames[14] do
      begin
        Number := 14;
        if FLanguage = nlSerbian then
          NameMasculine := 'cetrnaest'
        else
          NameMasculine := 'etrnaest';
        NameFeminine := ''
      end;

      with NumberNames[15] do
      begin
        Number := 15;
        NameMasculine := 'petanest';
        NameFeminine := ''
      end;

      with NumberNames[16] do
      begin
        Number := 16;
        if FLanguage = nlSerbian then
          NameMasculine := 'sesnaest'
        else
          NameMasculine := 'esnaest';
        NameFeminine := ''
      end;

      with NumberNames[17] do
      begin
        Number := 17;
        NameMasculine := 'sedamnaest';
        NameFeminine := ''
      end;

      with NumberNames[18] do
      begin
        Number := 18;
        NameMasculine := 'osamnaest';
        NameFeminine := ''
      end;

      with NumberNames[19] do
      begin
        Number := 19;
        NameMasculine := 'devetnaest';
        NameFeminine := ''
      end;

      with NumberNames[20] do
      begin
        Number := 20;
        NameMasculine := 'dvadeset';
        NameFeminine := ''
      end;

      with NumberNames[21] do
      begin
        Number := 30;
        NameMasculine := 'trideset';
        NameFeminine := ''
      end;

      with NumberNames[22] do
      begin
        Number := 40;
        if FLanguage = nlSerbian then
          NameMasculine := 'cetrdeset'
        else
          NameMasculine := 'etrdeset';
        NameFeminine := ''
      end;

      with NumberNames[23] do
      begin
        Number := 50;
        NameMasculine := 'pedeset';
        NameFeminine := ''
      end;

      with NumberNames[24] do
      begin
        Number := 60;
        if FLanguage = nlSerbian then
          NameMasculine := 'sezdeset'
        else
          NameMasculine := 'ezdeset';
        NameFeminine := ''
      end;

      with NumberNames[25] do
      begin
        Number := 70;
        NameMasculine := 'sedamdeset';
        NameFeminine := ''
      end;

      with NumberNames[26] do
      begin
        Number := 80;
        NameMasculine := 'osamdeset';
        NameFeminine := ''
      end;

      with NumberNames[27] do
      begin
        Number := 90;
        NameMasculine := 'devedeset';
        NameFeminine := ''
      end;

      with NumberNames[28] do
      begin
        Number := 100;
        NameMasculine := 'sto';
        NameFeminine := ''
      end;

      with NumberNames[29] do
      begin
        Number := 200;
        NameMasculine := 'dvesto';
        NameFeminine := 'dvesta'
      end;

      with NumberNames[30] do
      begin
        Number := 300;
        NameMasculine := 'tristo';
        NameFeminine := 'trista'
      end;

      with NumberNames[31] do
      begin
        Number := 400;
        if FLanguage = nlSerbian then
          NameMasculine := 'cetirsto'
        else
          NameMasculine := 'etirsto';
        NameFeminine := ''
      end;

      with NumberNames[32] do
      begin
        Number := 500;
        NameMasculine := 'petsto';
        NameFeminine := ''
      end;

      with NumberNames[33] do
      begin
        Number := 600;
        if FLanguage = nlSerbian then
          NameMasculine := 'sesto'
        else
          NameMasculine := 'esto';
        NameFeminine := ''
      end;

      with NumberNames[34] do
      begin
        Number := 700;
        NameMasculine := 'sedamsto';
        NameFeminine := ''
      end;

      with NumberNames[35] do
      begin
        Number := 800;
        NameMasculine := 'osamsto';
        NameFeminine := ''
      end;

      with NumberNames[36] do
      begin
        Number := 900;
        NameMasculine := 'devetsto';
        NameFeminine := ''
      end;

      // Zatim inicijalizujemo redove.

      SetLength(NumberOrders,6);

      with NumberOrders[0] do
      begin
        Order := 3;
        Gender := gFeminine;
        Name1Ignored := 'hiljadu';
        Name1 := 'hiljada';
        Name2 := 'hiljade';
        Name5 := 'hiljada';
      end;

      with NumberOrders[1] do
      begin
        Order := 6;
        Gender := gMasculine;
        Name1Ignored := 'milion';
        Name1 := 'milion';
        Name2 := 'miliona';
        Name5 := 'miliona';
      end;

      with NumberOrders[2] do
      begin
        Order := 9;
        Gender := gFeminine;
        Name1Ignored := 'milijardu';
        Name1 := 'milijarda';
        Name2 := 'milijarde';
        Name5 := 'milijardi';
      end;

      with NumberOrders[3] do
      begin
        Order := 12;
        Gender := gMasculine;
        Name1Ignored := 'bilion';
        Name1 := 'bilion';
        Name2 := 'biliona';
        Name5 := 'biliona';
      end;

      with NumberOrders[4] do
      begin
        Order := 15;
        Gender := gFeminine;
        Name1Ignored := 'bilijardu';
        Name1 := 'bilijarda';
        Name2 := 'bilijarde';
        Name5 := 'bilijardi';
      end;

      with NumberOrders[5] do
      begin
        Order := 18;
        Gender := gMasculine;
        Name1Ignored := 'trilion';
        Name1 := 'trilion';
        Name2 := 'triliona';
        Name5 := 'triliona';
      end;

    end
  end
end;

function TNumberToText.GetText: string;
var
  Num,Fraction: Int64;
  FractionExt (*,Log10Tmp*): Extended;
  FractionPos: integer;
begin

  Num := Round(Int(FNumber));

  FractionExt := Abs(Frac(FNumber));
  FractionPos := 0;
  while (FractionExt > 0) and ((FractionExt * 10) - 1 < 0.000000000001) do // 12 zeroes.
  begin
    FractionExt := FractionExt * 10;
    Inc(FractionPos)
  end;

  Fraction := Round(FractionExt * 1000000000000); // 12 zeroes.
  Fraction := TrimRightZeros(Fraction);

  result := NumberToText(
    NumberNames,
    NumberOrders,
    FMinusName,
    Num,
    FDecimalSeparator,
    Fraction,
    FractionPos,
    FSeparator,
    FIgnoreOne)

end;

procedure TNumberToText.SetText(const Value: string);
begin
  raise Exception.Create('TNumberToText.SetText() Unsupported feature.');
end;



{ TNumberToRoman }

function TNumberToRoman.GetRomanNumber: string;
begin
  result := CardToRoman(FNumber)
end;

procedure TNumberToRoman.SetRomanNumber(const Value: string);
begin
  raise Exception.Create('TNumberToRoman.SetRomanNumber() Unsupported feature.');
end;



end.
